/* cfp.c: Classifier with Feature Partitioning */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include "cfp.h"

int No_Features, No_Classes;
int No_Trainers, No_Testers, Train_correct, Test_correct;
float weight[MAX_FEATURES], gen_limit[MAX_FEATURES];
int linear[MAX_FEATURES];
float delta=0.0;
int LEARN_WEIGHTS;
int className[MAX_CLASSES], classIndex[MAX_CLASSES];
float vote[MAX_CLASSES];

INSTANCE trainData[MAX_DATA], testData[MAX_DATA];

char Domain[MAX_NAME], domainFile[MAX_NAME],
     trainDataFile[MAX_NAME], testDataFile[MAX_NAME],
     logFile[MAX_NAME];

int verbosity=0;
int testMode=0;

FILE *fopen(), *resultf, *logf;

struct timeval *tr_start, *ts_start, *ts_end;
struct timezone *tz;
int tr_time, ts_time;

/****************************************************/

parse_options(int argc, char *argv[])
{
  strcpy(Domain, argv[1]);
  strcpy(domainFile, Domain);
  strcat(domainFile, ".cfp");
  strcpy(trainDataFile, Domain);
  strcat(trainDataFile, ".train");
  strcpy(testDataFile, Domain);
  strcat(testDataFile, ".test");
  if (argc >= 3 && (strcmp(argv[2], "-v")==0)) {
    strcpy(logFile, Domain);
    strcat(logFile, ".log");
    if (argc == 3) verbosity = 1; /* default is the least verbosity */
    else sscanf(argv[3], "%d", &verbosity);
  }
}

void initialize()
{
  FEATURE f;

  No_Classes = No_Features = No_Testers = No_Testers = 0;
  read_domain(domainFile); 
  read_data(testDataFile, testData, &No_Testers);
  read_data(trainDataFile, trainData, &No_Trainers);
  initialize_partitions ();
  if (verbosity>0) {
    logf = fopen(logFile,"a"); 
    fprintf(logf, "Started logging with verbosity %d.\n", verbosity);
  }
  if (LEARN_WEIGHTS)
    for (f=0; f < No_Features; f++) weight[f] = 1.0;  
}

CLASS prediction_on_feature (f, value)
FEATURE f;
VALUE value;
{
  SEGMENT *seg, *s;

  seg = search (f, value);

  if (value < seg->upper)       /* in a range partition */
    return(seg->class);
  
  /* return the class of the segment with maximum repr */
  for (s = seg->right; (s->right !=NULL && s->lower <= value); s = s->right) 
    if (s->repr > seg->repr)
      seg = s;
  return (seg->class);
}

int classification (inst)
INSTANCE inst;
{
  CLASS c, pof, winner;
  FEATURE f;

  for (c=0; c <= No_Classes; c++) vote[c] = 0.0;

  for (f=0; f < No_Features; f++) 

    if (inst.status[f] == KNOWN) {
      pof = prediction_on_feature (f, inst.value[f]);
      vote[pof] = vote[pof] + weight[f];
    if ((verbosity>0) && testMode) fprintf(logf,"%d ", className[pof]);
    }
    else if ((verbosity>0) && testMode) fprintf(logf, "? ");

  /* return the class that received maximum votes */
  winner = 1;
  for (c=2; c <= No_Classes; c++) 
    if (vote[c] > vote[winner]) 
      winner = c;
  if (vote[winner] == 0.0) 
    winner = UNDETERMINED;
  if ((verbosity>0) && testMode) {
    if (winner == UNDETERMINED) fprintf(logf, "> ?\n");
    else fprintf(logf, "> %d\n", className[winner]);
  }
  return(winner);
}

float train()
{
  int i, f, pof; 
  int train_correct = 0;

  for (i=0; i < No_Trainers; i++) {
    if (classification(trainData[i]) == trainData[i].class)
      train_correct++;

    for (f=0; f < No_Features; f++) 
      if (trainData[i].status[f] == KNOWN) {
	if (LEARN_WEIGHTS) {
	  pof = prediction_on_feature (f, trainData[i].value[f]);
	  if (pof == trainData[i].class) 
	    weight[f] *= (1+delta);
	  else
	    weight[f] *= (1-delta);
	}
	update_feature_partitioning (f, trainData[i]);
      }
  }
  return((float)train_correct/No_Trainers);
}

float test()
{
  int i, test_correct = 0;
  CLASS c;

  testMode=1;
  for (i=0; i<No_Testers; i++) {
    if (verbosity>1) show_instance (testData[i]);
    c = classification(testData[i]);
    if (c == testData[i].class)
      test_correct++;
  }
  testMode=0;
  return ((float)test_correct/No_Testers);
}
	
float cfp()
{
  float train_acc, test_acc;

  gettimeofday(tr_start, tz);

  train_acc = train(); 
 
  gettimeofday(ts_start, tz);

  if (verbosity>2) {
    fprintf (logf, "Training is finished.\n");  
    show_partitions(); 
    show_weights();
  }
  if (verbosity>1) {
    fprintf (logf, "Total number of segments: %d\n", segment_count());
    fprintf (logf, "Train accucary: %g\n", train_acc);
  }

  test_acc = test();

  gettimeofday(ts_end, tz);

  if (verbosity>1) 
    fprintf (logf, "Test accuracy: %g\n", test_acc);

  return (test_acc);
}
  
main(argc, argv)
int argc;
char *argv[];
{

  float accuracy;
  int f;

  tr_start = (struct timeval * ) malloc(sizeof(struct timeval));
  ts_start = (struct timeval * ) malloc(sizeof(struct timeval));
  ts_end = (struct timeval * ) malloc(sizeof(struct timeval));
  tz    = (struct timezone * ) malloc(sizeof(struct timezone)); 

  if (argc == 1) {
    printf("cfp domain [-v [1|2|3|4]]\n");
    exit(1);
  }
  else
    parse_options(argc,argv);

  initialize();

  if (verbosity>3) {
    fprintf (logf, "Trainging data:\n");    
    show_data(trainData, No_Trainers); 
    fprintf(logf, "Test data\n");  
    show_data(testData, No_Testers); 
  }
  if (verbosity>2) {
    fprintf(logf, "Trainers: %d, Testers: %d\n", No_Trainers, No_Testers);
    fprintf(logf, "Features: %d, Classes: %d\n", No_Features, No_Classes);
    fprintf (logf, "Features: ");
    for (f=0; f<No_Features; f++)
      if (linear[f]) fprintf (logf, "l ");
      else fprintf (logf, "n ");
    fprintf(logf, "\n");
    if (LEARN_WEIGHTS)
      fprintf(logf, "Learning weights with delta: %g\n", delta);
    else {
      fprintf (logf, "Weights: ");
      for (f=0; f<No_Features; f++) fprintf(logf, "%g ", weight[f]);
    }
    fprintf(logf, "Generalization limits: ");
    for (f=0; f<No_Features; f++) fprintf(logf, "%g ", gen_limit[f]);
    fprintf(logf, "\n");
  }

  accuracy = cfp(); 

  /* timing */
  tr_time = (ts_start->tv_sec * 1000 - tr_start->tv_sec * 1000) +
    (ts_start->tv_usec / 1000 - tr_start->tv_usec / 1000);
  ts_time = (ts_end->tv_sec * 1000 - ts_start->tv_sec * 1000) +
    (ts_end->tv_usec / 1000 - ts_start->tv_usec / 1000);
  
  resultf = fopen("result", "w");
  fprintf (resultf, "Accuracy: %6.2f Memory: %d segments. Training: %d msec, Testing: %d msec.\n", 
	   100.0 * accuracy, segment_count(), tr_time, ts_time);
  if (verbosity>1)
  fprintf (logf, "Accuracy: %6.2f Memory: %d segments. Training: %d msec, Testing: %d msec.\n", 
	   100.0 * accuracy, segment_count(), tr_time, ts_time);
  fclose(resultf);
  if (verbosity>0) fclose(logf);
  return 0;
}
